/* This file is part of the gf2x library.

   Copyright 2007, 2008, 2009
   Richard Brent, Pierrick Gaudry, Emmanuel Thome', Paul Zimmermann

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
   02111-1307, USA.
*/


/* This header file is unprotected on purpose */
#if TUNING == 1
#include "tuning_undefs_1.h"
#elif TUNING == 2
#include "tuning_undefs_2.h"
#elif TUNING == 3
#include "tuning_undefs_3.h"
#elif TUNING == 4
#include "tuning_undefs_4.h"
#elif TUNING == 5
#include "tuning_undefs_5.h"
#elif TUNING == 6
#include "tuning_undefs_6.h"
#elif TUNING == 7
#include "tuning_undefs_7.h"
#elif TUNING == 8
#include "tuning_undefs_8.h"
#elif TUNING == 9
#include "tuning_undefs_9.h"
#endif
